(    Title:  Tests for ttester-xf arrays
      File:  xfarray-test.fs
  Log File:  xftester.log
    Author:  David N. Williams
   License:  Public Domain
   Version:  1.4.3
   Revised:  January 18, 2021

The date above may reflect unlogged cosmetic changes.

This file tests the array factors in ttester-xf.  It should be tested
with an "original" version of ttester, one without explicit array words,
as a partial bootstrap for testing ttester-xf with itself.
)
decimal

\ debugging
[UNDEFINED] \\ [IF]
: \\  ( -- )  -1 parse 2drop BEGIN refill 0= UNTIL ;
[THEN]

: .#  ( n -- )  cr ." #" . cr ;

s" ttester.fs" INCLUDED
\ false CONSTANT USE-TERMINAL-COLORS
s" tester-display.fs" INCLUDED
true VERBOSE !

\ before the first displayed line
?.FIRST-CR


\ *** TTESTER ERROR DISPLAY

VARIABLE #errors    0 #errors !
:noname  ( c-addr u -- )
\ Display an error message followed by the line that had the
\ error.
  red-text 1 #errors +! ERROR1 normal-text ; ERROR-XT !

: ?.errors  ( -- )
  VERBOSE @
  IF blue-text ." #ERRORS: " normal-text #errors @ . THEN ;

\ *** XT-ARRAY WORDS
(
The first array element has index zero.
)

\ overflow checking to be done elsewhere
: XT-A!  ( x_[n-1] ... x_0 +n addr -- )
  OVER 1+ CELLS OVER + SWAP DO I ! 1 CELLS +LOOP ;

: XT-A@  ( addr -- x_[n-1] ... x_0 +n )
  DUP @ CELLS OVER + DO I @ -1 CELLS +LOOP ;

[UNDEFINED] CELL- [IF]
  : CELL-  ( addr -- addr-cell )  [ 1 CELLS ] LITERAL - ;
[THEN]

: XT-A=  ( addr1 addr2 -- 0 true | 0 false | index+1 false )
(
Returns:
         0 true:  equal size and elements
        0 false:  sizes unequal
  index+1 false:  sizes equal, deepest unequal elements at index
)
  2DUP @ SWAP @ <>
  IF 2DROP 0 FALSE EXIT THEN  \ unequal size
  2>R 0 TRUE 2R>
  ( 0 true addr1 addr2)
  DUP @ ( count) >R
  R@ CELLS + SWAP R@ CELLS + 0 R>
  ( 0 true addr2_last addr1_last 0 last_index+1)
  ?DO
    2DUP @ SWAP @ <>
    IF  \ unequal elements
      2DROP ( 0 true) 2DROP I FALSE 0 0 LEAVE THEN
    CELL- SWAP CELL- ( addr2' addr1')
  -1 +LOOP
  ( [0 true | i false] x1 x2) 2DROP ;

\ *** XT-ARRAY TESTS

TESTING  XT-A= XT-A@ XT-A!

CREATE []         0 ,
CREATE [0]        1 ,  0 ,
CREATE [10]       1 , 10 ,
CREATE [10,11]    2 , 10 , 11 ,
CREATE [10,1]     2 , 10 ,  1 ,
CREATE [0,11]     2 ,  0 , 11 ,
CREATE [0,1]      2 ,  0 ,  1 ,
CREATE [10,11,12] 3 , 10 , 11 , 12 ,
CREATE [1,11,12]  3 ,  1 , 11 , 12 ,
CREATE [1,1,12]   3 ,  1 ,  1 , 12 ,
CREATE [1,11,2]   3 ,  1 , 11 ,  2 ,

\ XT-A= Returns
\         0 true:  equal size and elements
\        0 false:  sizes unequal
\  index+1 false:  sizes equal, deepest unequal elements at index

\ different length
T{ []    [10,11,12] XT-A= -> 0 false }T
T{ [10]  [10,1]     XT-A= -> 0 false }T
T{ [10]  [10,11,12] XT-A= -> 0 false }T

\ last different at index 0
T{ [0]        [10]      XT-A= -> 1 false }T
T{ [10,1]     [0,1]     XT-A= -> 1 false }T
T{ [10,11,12] [1,11,12] XT-A= -> 1 false }T

\ last different at index 1
T{ [10,1]     [10,11]   XT-A= -> 2 false }T
T{ [10,1]     [0,11]    XT-A= -> 2 false }T
T{ [1,1,12]   [1,11,12] XT-A= -> 2 false }T
T{ [10,11,12] [1,1,12]  XT-A= -> 2 false }T

\ last different at index 2
T{ [1,11,12] [1,11,2] XT-A= -> 3 false }T
T{ [1,1,12]  [1,11,2] XT-A= -> 3 false }T

\ the same
T{ []         []         XT-A= -> 0 true }T
T{ [10]       [10]       XT-A= -> 0 true }T
T{ [10,11,12] [10,11,12] XT-A= -> 0 true }T

T{ []         XT-A@ -> 0 }T
T{ [10]       XT-A@ -> 10 1 }T
T{ [10,11]    XT-A@ -> 11 10 2 }T
T{ [10,11,12] XT-A@ -> 12 11 10 3 }T

create max3array 4 cells allot
T{ 3  2  1 3 max3array XT-A! max3array XT-A@ -> 3  2  1 3 }T
T{   11 10 2 max3array XT-A! max3array XT-A@ -> 11 10 2 }T
T{       0 1 max3array XT-A! max3array XT-A@ -> 0 1 }T
T{         0 max3array XT-A! max3array XT-A@ -> 0 }T
\ xarray overwrite?
T{ -1 -1 -1 3 max3array XT-A! -> }T
T{     0  0 2 max3array XT-A! -> }T
T{ max3array 2 cells + 2@ -> -1 0 }T

HAS-FLOATING 0= [IF]
?." Skipping floating-point array tests.
[ELSE]
DEPTH 1E DEPTH 1- FDROP = 0=
[IF] .( FLOATING-STACK REQUIRED) ABORT [THEN]

\ *** FT-ARRAY WORDS
(
The first element has index zero.
)

: FT-TEST=  FCONF= ;

\ Overflow checking to be done elsewhere.
: FT-A!   ( +n addr -- ) ( f: r_[n-1] ... r_0 -- )
  2DUP ! CELL+ FALIGNED SWAP ( f-addr n) DUP
  IF   FLOATS OVER + SWAP
       DO I F! 1 FLOATS +LOOP 
  ELSE 2DROP THEN ;

: FT-A@   ( addr -- +n ) ( f: -- r_[n-1] ... r_0 )
  DUP @ DUP >R ( addr n r: n)
  IF   CELL+ FALIGNED R@ 1- FLOATS OVER + 
       DO I F@ -1 FLOATS +LOOP
  ELSE DROP THEN R> ;

[UNDEFINED] FLOAT- [IF]
  : FLOAT-  ( addr -- addr-float )  [ 1 FLOATS ] LITERAL - ;
[THEN]

: FT-A=  ( addr1 addr2 -- 0 true | 0 false | index+1 false )
(
Returns:
         0 true:  equal size and elements
        0 false:  sizes unequal
  index+1 false:  sizes equal, deepest unequal elements at index
)
  2DUP @ SWAP @ <>
  IF 2DROP 0 FALSE EXIT THEN          \ unequal size
  DUP @ 0= IF 2DROP 0 TRUE EXIT THEN  \ equal size and empty
  2>R 0 TRUE 2R>
  ( 0 true addr1 addr2)
  DUP @ ( count) 1- ( last_index) >R
  CELL+ FALIGNED R@ FLOATS + SWAP
  CELL+ FALIGNED R@ FLOATS + 0 R>
  ( 0 true f-addr2_last f-addr1_last 0 last_index)
  DO
    2DUP F@ F@ FT-TEST= 0=
    IF  \ unequal elements
      2DROP ( 0 true) 2DROP I 1+ FALSE 0 0 LEAVE THEN
    FLOAT- SWAP FLOAT- ( f-addr2' f-addr1')
  -1 +LOOP
  ( [0 true | i+1 false] x1 x2) 2DROP ;

\ *** FT-ARRAY TESTS

TESTING  FT-A= FT-A= FT-A@ FT-A!

SET-EXACT

: FARRAY  ( +n -- ) ( f: x_[n-1] ... x_0 -- )
  CREATE dup , falign here >r floats allot here r>
  ?DO i f! 1 floats +LOOP ;

0             FARRAY f[]
1 0E          FARRAY f[0]
1 10E         FARRAY f[10]
2 11E 10E     FARRAY f[10,11]
2  1E 10E     FARRAY f[10,1]
2 11E  0E     FARRAY f[0,11]
2  1E  0E     FARRAY f[0,1]
3 12E 11E 10E FARRAY f[10,11,12]
3 12E 11E  1E FARRAY f[1,11,12]
3 12E  1E  1E FARRAY f[1,1,12]
3  2E 11E  1E FARRAY f[1,11,2]

0             FARRAY g[]
1 0E          FARRAY g[0]
1 10E         FARRAY g[10]
2 11E 10E     FARRAY g[10,11]
2  1E 10E     FARRAY g[10,1]
2 11E  0E     FARRAY g[0,11]
2  1E  0E     FARRAY g[0,1]
3 12E 11E 10E FARRAY g[10,11,12]
3 12E 11E  1E FARRAY g[1,11,12]
3 12E  1E  1E FARRAY g[1,1,12]
3  2E 11E  1E FARRAY g[1,11,2]

T{ f[] @ -> 0 }T
T{ f[10] @ g[10] cell+ faligned f@ -> 1 10E }T
T{ f[10,11] @ g[10,11] cell+ faligned dup f@ float+ f@ -> 2 10e 11e }T

\ different length
T{ f[]    g[10,11,12] FT-A= -> 0 false }T
T{ f[10]  g[10,1]     FT-A= -> 0 false }T
T{ f[10]  g[10,11,12] FT-A= -> 0 false }T

\ FT-A= Returns
\        0 true:  equal size and elements
\       0 false:  sizes unequal
\ index+1 false:  sizes equal, deepest unequal elements at index

\ different length
T{ f[]    g[10,11,12] FT-A= -> 0 false }T
T{ f[10]  g[10,1]     FT-A= -> 0 false }T
T{ f[10]  g[10,11,12] FT-A= -> 0 false }T

\ last different at index 0
T{ f[0]        g[10]      FT-A= -> 1 false }T
T{ f[10,1]     g[0,1]     FT-A= -> 1 false }T
T{ f[10,11,12] g[1,11,12] FT-A= -> 1 false }T

\ last different at index 1
T{ f[10,1]     g[10,11]   FT-A= -> 2 false }T
T{ f[10,1]     g[0,11]    FT-A= -> 2 false }T
T{ f[1,1,12]   g[1,11,12] FT-A= -> 2 false }T
T{ f[10,11,12] g[1,1,12]  FT-A= -> 2 false }T

\ last different at index 2
T{ f[1,11,12] g[1,11,2] FT-A= -> 3 false }T
T{ f[1,1,12]  g[1,11,2] FT-A= -> 3 false }T

\ the same
T{ f[]         g[]         FT-A= -> 0 true }T
T{ f[10]       g[10]       FT-A= -> 0 true }T
T{ f[10,11,12] g[10,11,12] FT-A= -> 0 true }T

T{ f[]         FT-A@ -> false }T
T{ f[10]       FT-A@ -> 10E 1 }T
T{ f[10,11]    FT-A@ -> 11E 10E 2 }T
T{ f[10,11,12] FT-A@ -> 12E 11E 10E 3 }T

here 1 cells allot falign 3 floats allot CONSTANT max3farray
T{ 3E  2E  1E 3 max3farray FT-A! max3farray FT-A@ ->  3E  2E  1E 3 }T
T{    11E 10E 2 max3farray FT-A! max3farray FT-A@ -> 11E 10E 2 }T
T{         0E 1 max3farray FT-A! max3farray FT-A@ ->  0E 1 }T
T{            0 max3farray FT-A! max3farray FT-A@ ->  0 }T
\ farray overwrite?
T{ -1E -1E -1E 3 max3farray FT-A! -> }T
T{      0E  0E 2 max3farray FT-A! -> }T
T{ max3farray cell+ faligned float+ dup f@ float+ f@ -> 0E -1E }T

[THEN]  \ HAS-FLOATING

true VERBOSE !
\ ?." Host is " HOST-S ?.TYPE ?.CR
?.errors GFORTH-HOST [IF] ?.cr [THEN]
